import os
import json
import shutil
from typing import Tuple
import yaml
import gzip
import logging
import requests
import random
from shutil import copyfileobj
from datetime import datetime
from time import time

FH_DIR_PATH = os.path.dirname(os.path.realpath(__file__))
CONTEXT_PATH = os.path.join(FH_DIR_PATH, "..", "data", "context.json")


class FileHandler:
    """A class to manage cache and log files.

    Folder structure:
    Storage/
        Cache/
            Exampleid/
                messages.json.gz
                metadata.yaml
            ...
        Logs/
        Exports/
    """

    def __init__(
        self,
        storage_path,
        cache_fname="Cache",
        log_fname="Logs",
        export_fname="Exports",
        message_fname="messages.json",
        metadata_fname="metadata.yaml",
        thumbnail_fname="thumbnail.png",
        graph_fname="graph.png",
        wordcloud_fname="wordcloud.jpg",
    ):
        self.storage_path = storage_path
        self.cache_path = os.path.join(self.storage_path, cache_fname)
        self.log_path = os.path.join(self.storage_path, log_fname)
        self.export_path = os.path.join(self.storage_path, export_fname)
        self.message_fname = message_fname
        self.metadata_fname = metadata_fname
        self.thumbnail_fname = thumbnail_fname
        self.graph_fname = graph_fname
        self.wordcloud_fname = wordcloud_fname

        # create_logger is seperately implemented to prevent circular imports
        self.logger = self._create_logger(__file__)

        self.sid_path = None

    def __repr__(self) -> str:
        return "Storing files into " + self.storage_path

    def _create_logger(self, name, def_level=logging.ERROR, level=logging.DEBUG):
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)

        format = f"%(asctime)s:%(module)s[%(lineno)d]:%(levelname)s:%(message)s"

        if not os.path.exists(self.log_path):
            os.makedirs(self.log_path)

        logging.basicConfig(
            level=def_level,
            format=format,
            filename=os.path.join(self.log_path, self._get_logname()),
            encoding="utf-8",
        )
        console = logging.StreamHandler()
        console.setLevel(def_level)
        console.setFormatter(logging.Formatter(format))
        logging.getLogger(name).addHandler(console)
        logger = logging.getLogger(name)
        logger.setLevel(level)
        return logger

    def _get_logname(self) -> str:
        """Gets log name in Y-M-Wn format where n is week number, starts from 0
        Example: 2021-06-W0"""
        weekno = (
            datetime.today().isocalendar()[1]
            - datetime.today().replace(day=1).isocalendar()[1]
        )
        return datetime.today().strftime("%Y-%m-W") + str(weekno) + ".log"

    def delete_file(self, path):
        try:
            os.remove(path)
            self.logger.debug(f"{path} file removed")
        except FileNotFoundError:
            self.logger.warning(f"{path} could not be found")
        except PermissionError:
            self.logger.error(f"Access is denied to {path}. Try running in administrator mode.")
        except Exception as e:
            self.logger.critical(f"Could not remove {path} - {e.__class__.__name__}:{e}")

    def delete_dir(self, path):
        try:
            shutil.rmtree(path)
            self.logger.debug(f"{path} folder removed")
        except FileNotFoundError:
            self.logger.warning(f"{path} could not be found")
        except PermissionError:
            self.logger.error(f"Access is denied to {path}. Try running in administrator mode.")
        except Exception as e:
            self.logger.critical(f"Could not remove {path} - {e.__class__.__name__}:{e}")

    def create_dir_if_not_exists(self, path):
        if not os.path.exists(path):
            try:
                os.makedirs(path)
                self.logger.debug(f"'{path}' created")
            except PermissionError as e:
                print(
                    f"{e}\nTry another path or re-run the program in administrator mode."
                )
                self.logger.error(f"Permission denied to '{path}'")

    def create_cache_dir(self, stream_id):
        self.sid_path = os.path.join(self.cache_path, stream_id)
        self.create_dir_if_not_exists(self.sid_path)

    def cache_messages(self, message_dict):
        self.logger.info("Caching messages")
        fpath = os.path.join(self.sid_path, self.message_fname)
        try:
            with open(fpath, "w+", encoding="utf-8") as f:
                f.write(json.dumps(message_dict, ensure_ascii=False, indent=4))
        except Exception as e:
            self.delete_file(fpath)
            raise RuntimeError(f"Could not cache messages: {e.__class__.__name__}:{e}")

        self._compress_file(fpath)

    def cache_metadata(self, metadata_dict):
        self.logger.info("Caching metadata")
        fpath = os.path.join(self.sid_path, self.metadata_fname)
        try:
            with open(fpath, "w+", encoding="utf-8") as file:
                yaml.dump(
                    metadata_dict, file, default_flow_style=False, allow_unicode=True
                )
        except Exception as e:
            self.delete_file(fpath)
            raise RuntimeError(f"Could not cache metadata: {e.__class__.__name__}:{e}")

    def cache_thumbnail(self, url):
        """Alias for `download_thumbnail`"""
        self.download_thumbnail(url)

    def download_thumbnail(self, url, destination):
        self.logger.info("Downloading thumbnail")
        try:
            response = requests.get(url)
            with open(destination, "wb") as f:
                f.write(response.content)
        except Exception as e:
            self.delete_file(destination)
            raise RuntimeError(
                f"Could not download thumbnail: {e.__class__.__name__}:{e}"
            )

    def read_messages(self):
        """Reads cached messages.
        Returns a dict."""
        fpath = os.path.join(self.sid_path, self.message_fname)
        self._decompress_file(fpath)
        with open(fpath, "r", encoding="utf-8") as f:
            data = json.load(f)
        self.logger.info("Read messages")
        self._compress_file(fpath)
        return data

    def read_metadata(self):
        """Reads cached messages.
        Returns a dict."""
        fpath = os.path.join(self.sid_path, self.metadata_fname)
        self.logger.info("Read metadata")
        return yaml.load(open(fpath, "r", encoding="utf-8"), Loader=yaml.Loader)

    def _compress_file(self, jsonpath):
        """Compresses a json file with gzip"""
        try:
            with open(jsonpath, "rb") as f_in:
                with gzip.open(jsonpath + ".gz", "wb") as f_out:
                    copyfileobj(f_in, f_out)
            os.remove(jsonpath)
        except Exception as e:
            os.remove(jsonpath + ".gz")
            self.logger.critical(e)
            raise e
        self.logger.info(f"{jsonpath} compressed")

    def _decompress_file(self, jsonpath):
        """Decompresses a json gzip file"""
        try:
            with gzip.open(jsonpath + ".gz", "rb") as f_in:
                with open(jsonpath, "wb") as f_out:
                    copyfileobj(f_in, f_out)
            os.remove(jsonpath + ".gz")
        except Exception as e:
            os.remove(jsonpath)
            self.logger.critical(e)
            raise e
        self.logger.info(f"{jsonpath} decompressed")

    def clear_cache(self, cache_deletion_algorithm=None, delete_root_folder=True):
        """Clears cached files according to cache deletion
            algorithm. Default behavior is deleting the
            cache of the current session.

        Args:
            cache_deletion_algorithm (str|None, optional): Algorithm to
                delete cached files. Options are as follows:
                    - lru: Deletes least recently
                        used cache.
                    - mru: Deletes most recently
                        used cache.
                    - fifo: Deletes oldest cache.
                    - rr: Deletes random cache. (uhh...)
                If set to None, deletes cache of the current session.
                Defaults to None.

            delete_root_folder (bool, optional): Delete the folder itself too.
                Otherwise only deletes contents in the folder. Defaults to True.
        """

        self.logger.info("Clearing cache")
        self.logger.debug(f"{cache_deletion_algorithm=}")
        self.logger.debug(f"{delete_root_folder=}")

        if not cache_deletion_algorithm:
            if self.sid_path:
                try:
                    shutil.rmtree(self.sid_path)
                    self.logger.debug(f"Deleted '{self.sid_path}'")
                    if not delete_root_folder:
                        os.makedirs(self.sid_path)
                except Exception as e:
                    self.logger.error(e)
            return

        if cache_deletion_algorithm == "mru":
            dir_to_delete = self.most_recently_used_folder(self.cache_path)
        elif cache_deletion_algorithm == "lru":
            dir_to_delete = self.least_recently_used_folder(self.cache_path)
        elif cache_deletion_algorithm == "fifo":
            dir_to_delete = self.oldest_folder(self.cache_path)
        elif cache_deletion_algorithm == "rr":
            dir_to_delete = self.random_folder(self.cache_path)
        else:
            self.logger.error(
                "Invalid deletion algorithm: {}".format(cache_deletion_algorithm)
            )
            raise ValueError(
                "Invalid deletion algorithm: {}".format(cache_deletion_algorithm)
            )

        path = os.path.join(self.cache_path, dir_to_delete)
        try:
            shutil.rmtree(path)
            self.logger.debug(f"Deleted '{self.sid_path}'")
            if not delete_root_folder:
                os.makedirs(path)
        except Exception as e:
            self.logger.error(e)

    def check_integrity(self, cache_path=None, autofix=False) -> Tuple[list, list]:
        """Checks integrity of the cached files.
        Note that it detects files by their names, not content.

        Args:
            cache_path (str|None, optional): Path to the cache files
                of a stream. Defaults to None, which sets the path to
                the current stream id.

            autofix (bool, optional): Attempts to automatically fix issues
                by deleting unnecessary files and compressing uncompressed
                messages. Missing files must be dealt manually.
                Defaults to False.

        Returns:
            Tuple[list, list]: Lists of unnecesary and missing files
        """

        self.logger.info("Checking cache integrity")
        self.logger.debug(f"{cache_path=}")
        self.logger.debug(f"{autofix=}")

        if not cache_path:
            cache_path = self.sid_path

        files = self.get_filenames(cache_path, show_extension=True)
        necessary_files = [
            self.message_fname + ".gz",
            self.metadata_fname,
        ]
        unnecesary_files = list(set(files) - set(necessary_files))
        missing_files = list(set(necessary_files) - set(files))

        self.logger.debug(f"{unnecesary_files=}")
        self.logger.debug(f"{missing_files=}")

        if autofix:
            for folder in os.listdir(self.cache_path):
                full_path = os.path.join(self.cache_path, folder)
                try:
                    for _, _, files in os.walk(full_path):
                        pass
                    if os.stat(full_path).st_size == 0 and len(files) == 0:
                        self.delete_dir(full_path)
                except AttributeError:
                    continue
            for file in unnecesary_files:
                # it might be a json file that is not compressed
                if file == self.message_fname:
                    self._compress_file(os.path.join(cache_path, file))
                    missing_files.remove(file + ".gz")
                    continue
                self.delete_file(os.path.join(cache_path, file))
            unnecesary_files = []

        return missing_files, unnecesary_files

    def get_filenames(self, path, show_extension=False) -> list[str]:
        """Returns file names in a path"""

        self.logger.info(f"Finding file names in {path}")
        if not os.path.exists(path):
            self.logger.error("Path doesn't exist")
            return []

        try:
            if show_extension:
                return [fname for fname in os.listdir(path)]
            return [fname.split(os.extsep)[0] for fname in os.listdir(path)]
        except Exception as e:
            self.logger.error(e)

    def get_foldernames(self, path) -> list[str]:
        """Returns folder names in a path"""

        self.logger.info(f"Finding folder names in {path}")
        try:
            return next(os.walk(path))[1]
        except Exception as e:
            self.logger.error(e)
            raise e

    def _creation_time_in_days(self, path) -> int:
        """Returns difference between the creation time and
        the current time of the file in days"""

        if os.path.isfile(path):
            ctime = os.path.getctime(path)
            days = datetime.fromtimestamp(int(time() - ctime)).strftime("%d")
            return int(days)
        self.logger.debug(f"{path} is not a file")
        return 0

    def _delete_old_files(self, folder_path, time_limit_in_days):
        """Delete files in a folder older than the time limit."""

        for name in os.listdir(folder_path):
            fpath = os.path.join(folder_path, name)
            if self._creation_time_in_days(fpath) >= time_limit_in_days:
                self.delete_file(fpath)

    def file_amount(self, folder_path) -> int:
        """Returns file amount in a folder"""

        _, _, files = next(os.walk(folder_path))
        self.logger.debug(f"File amount in {folder_path} is {len(files)}")
        return len(files)

    def dir_amount(self, folder_path) -> int:
        """Returns folder amount in a folder"""

        _, dirs, _ = next(os.walk(folder_path))
        self.logger.debug(f"Folder amount in {folder_path} is {len(dirs)}")
        return len(dirs)

    def least_recently_used_folder(self, fpath) -> str:
        """Returns least recenty used folder in path
        by checking last access time of the
        message file inside the folder"""

        least_recently_used_folder = None
        last_access_time = time()

        for root, dirs, _ in os.walk(fpath):
            for dname in dirs:
                if os.path.join(root, dname) == self.sid_path:
                    continue
                try:
                    path = os.path.join(root, dname, self.message_fname + ".gz")
                except:
                    try:
                        self.check_integrity(
                            cache_path=os.path.join(root, dname), autofix=True
                        )
                        path = os.path.join(root, dname, self.message_fname + ".gz")
                    except:
                        msg = "Message cache couldn't be found"
                        self.logger.error(msg)
                        raise Exception(msg)
                if os.path.getatime(path) < last_access_time:
                    last_access_time = os.path.getatime(path)
                    least_recently_used_folder = dname
        self.logger.debug(
            "Least recently used id: {} ({})".format(
                least_recently_used_folder, last_access_time
            )
        )
        return least_recently_used_folder

    def most_recently_used_folder(self, fpath) -> str:
        """Returns most recenty used folder in path
        by checking last access time of the
        message file inside the folder"""

        most_recently_used_folder = None
        last_access_time = 0

        for root, dirs, _ in os.walk(fpath):
            for dname in dirs:
                if os.path.join(root, dname) == self.sid_path:
                    continue
                try:
                    path = os.path.join(root, dname, self.message_fname + ".gz")
                except:
                    try:
                        self.check_integrity(
                            cache_path=os.path.join(root, dname), autofix=True
                        )
                        path = os.path.join(root, dname, self.message_fname + ".gz")
                    except:
                        msg = "Message cache couldn't be found"
                        self.logger.error(msg)
                        raise Exception(msg)
                if os.path.getatime(path) > last_access_time:
                    last_access_time = os.path.getatime(path)
                    most_recently_used_folder = dname
        self.logger.debug(
            "Most recently used id: {} ({})".format(
                most_recently_used_folder, last_access_time
            )
        )
        return most_recently_used_folder

    def oldest_folder(self, fpath) -> str:
        """Returns oldest folder in path"""

        oldest_folder = None
        creation_time = time()

        for root, dirs, _ in os.walk(fpath):
            for dname in dirs:
                if os.path.join(root, dname) == self.sid_path:
                    continue
                path = os.path.join(root, dname)
                if os.path.getctime(path) < creation_time:
                    creation_time = os.path.getctime(path)
                    oldest_folder = dname

        self.logger.debug(f"Oldest folder: {oldest_folder} ({creation_time})")
        return oldest_folder

    def random_folder(self, fpath) -> str:
        """Returns a random folder in cache path"""
        for root, dirs, _ in os.walk(fpath):
            if len(dirs) == 1:
                return dirs[0]
            while True:
                choice = random.choice(dirs)
                if os.path.join(root, choice) != self.sid_path:
                    break
            self.logger.debug(f"Random folder: {choice}")
            return choice

    def is_cached(self, sid=None) -> bool:
        """Returns False if all necessary files are absent"""

        self.logger.info("Checking if cached")
        self.logger.debug(f"{sid=}")

        if sid:
            path = os.path.join(self.cache_path, sid)
        else:
            path = self.sid_path

        files = self.get_filenames(path, show_extension=True)
        necessary_files = [
            self.message_fname + ".gz",
            self.metadata_fname,
        ]
        missing_files = list(set(necessary_files) - set(files))

        self.logger.debug(f"is_cached: {(not len(missing_files) == 2)}")
        return not len(missing_files) == 2

    def get_cached_ids(self) -> list[str]:
        """Returns list of cached ids"""
        return [f for f in self.get_foldernames(self.cache_path) if self.is_cached(f)]

    def add_context(self, reaction_to, phrases):
        """Add new context. Note that it does not check
        if the context already exists

        Args:
            reaction_to (str): What is the reaction for.
            phrases (list[dict]): Phrase list for the reaction.
                Each element should be a dictionary that contains:
                - phrase (str): Phrase that triggers the reaction.
                - is_exact (boolean): If the message should be exactly the same
                    or should only include it.

                example: `[{phrase:"lol", is_exact:False}]`
        """
        self.logger.info("Adding new context")
        self.logger.debug(f"{reaction_to=}")
        self.logger.debug(f"{phrases=}")

        new_context = {
            "reaction_to": reaction_to,
            "triggers": [
                {"phrase": phrase["phrase"], "is_exact": phrase["is_exact"]}
                for phrase in phrases
            ],
        }
        with open(CONTEXT_PATH, "r+", encoding="utf-8") as file:
            contexts = list(json.load(file))
            contexts.append(new_context)
            file.seek(0)
            file.write(json.dumps(contexts, indent=4, ensure_ascii=False))

    def remove_context(self, reaction_to):
        """Remove a context
        Args:
            reaction_to (str): Reaction to be deleted.
        """
        self.logger.info("Removing context")
        self.logger.debug(f"{reaction_to=}")

        with open(CONTEXT_PATH, "r", encoding="utf-8") as file:
            contexts = list(json.load(file))
        for i in range(len(contexts)):
            if contexts[i]["reaction_to"] == reaction_to:
                del contexts[i]
                break
        with open(CONTEXT_PATH, "w", encoding="utf-8") as file:
            file.write(json.dumps(contexts, indent=4, ensure_ascii=False))

    def open_cache_folder(self, sid):
        target_path = os.path.join(self.cache_path, sid)
        try:
            os.startfile(target_path)
            self.logger.info(f"Opened {target_path} in file explorer")
        except FileNotFoundError:
            self.logger.error(f"Could not find {target_path}")



import sys
import os
import argparse

import streamanalyser as sa

# a basic CLI to fulfill the core features


def parseargs():
    parser = argparse.ArgumentParser()
    parser.add_argument("id", help="id of the YouTube live stream")
    parser.add_argument(
        "-s",
        "--silent",
        action="store_true",
        help="make output silent",
    )
    parser.add_argument(
        "-l", "--limit", default=None, type=int, help="message limit to fetch"
    )
    parser.add_argument(
        "-t", "--top", default=None, type=int, help="option to return top n highlights"
    )
    parser.add_argument(
        "-if",
        "--intensity-filters",
        nargs="+",
        default=[],
        help="highlight intensity levels to filter out",
    )
    parser.add_argument(
        "-kf", "--keyword-filters", nargs="+", default=[], help="keywords to filter out"
    )
    parser.add_argument(
        "-inc",
        "--include-context",
        nargs="+",
        default=[],
        help="keywords to filter out",
    )
    parser.add_argument(
        "-exc",
        "--exclude-context",
        nargs="+",
        default=[],
        help="keywords to filter out",
    )
    parser.add_argument(
        "-kw",
        "--keyword-limit",
        default=4,
        type=int,
        help="keyword amount to get for each highlight",
    )
    parser.add_argument(
        "-w",
        "--window",
        default=30,
        type=int,
        help="window to calculate moving average of message frequency",
    )
    parser.add_argument(
        "-ho",
        "--highlight-output",
        default=None,
        type=str,
        help="print highlights. Options are 'detailed', 'summary' and 'url'",
    )
    parser.add_argument("-g", "--graph", action="store_true", help="show graph")
    parser.add_argument(
        "-sca", "--show-cached", action="store_true", help="show cached file ids"
    )
    parser.add_argument(
        "-exp",
        "--export",
        nargs="?",
        const="default",
        type=str,
        help="export data to a specified path, leave empty to use the default path",
    )
    parser.add_argument(
        "-efn",
        "--export-folder-name",
        default=None,
        type=str,
        help="sets export folder name",
    )
    parser.add_argument(
        "-fm",
        "--find-messages",
        nargs="?",
        const="",
        type=str,
        help="find messages containing the given phrase",
    )
    parser.add_argument(
        "-e",
        "--exact",
        action="store_true",
        required="--find-message" in sys.argv,
        help="option for matching phrase to be exactly the same with the message, not only a part it",
    )
    parser.add_argument(
        "-sc",
        "--strict-case",
        action="store_true",
        required="--find-message" in sys.argv,
        help="option to match the case exactly. must be used with --find-message option",
    )
    parser.add_argument(
        "-fum",
        "--find-user-messages",
        action="store_true",
        help="find messages the specified user made\nmust be used with --user-id or --username",
    )
    user_info = parser.add_mutually_exclusive_group()
    user_info.add_argument("--user-id", default=None, type=str, help="id of the user")
    user_info.add_argument(
        "--username", default=None, type=str, help="username of the user"
    )
    parser.add_argument(
        "-ns",
        "--no-sound",
        action="store_true",
        help="do not make a sound when the program is completed",
    )
    parser.add_argument(
        "--open-in-chrome",
        action="store_true",
        help="open top highlights in chrome in descending order. must be used with --top and --highlights options",
    )
    parser.add_argument(
        "--thumb-res-lvl",
        default=2,
        type=int,
        help="thumbnail resolution level (0-3) with 0 being lowest and 3 being highest",
    )
    parser.add_argument(
        "-dl",
        "--disable-logs",
        action="store_true",
        help="actions done withing the current session will not be logged",
    )
    parser.add_argument(
        "-ld",
        "--log-duration",
        default=15,
        type=int,
        help="how old a log file should be to get deleted (in days)",
    )
    parser.add_argument(
        "-r", "--reset", action="store_true", help="clear cache before analysing"
    )
    parser.add_argument(
        "-nc", "--not-cache", action="store_true", help="clear cache after analysing "
    )
    parser.add_argument(
        "-md",
        "--min-duration",
        default=15,
        type=int,
        help="minimum highlight duration in seconds",
    )
    parser.add_argument(
        "-oef",
        "--open-export-folder",
        action="store_true",
        help="open the export folder in file explorer after exporting",
    )
    parser.add_argument(
        "-wc", "--wordcloud", action="store_true", help="show word cloud"
    )
    parser.add_argument(
        "-wcs", "--wordcloud-scale", default=3, type=int, help="scale of the wordcloud"
    )
    user_info.add_argument(
        "--yt-api-key", default="", type=str, help="youtube api key"
    )
    return parser.parse_args()


def main():
    args = parseargs()

    analyser = sa.StreamAnalyser(
        args.id,
        msglimit=args.limit,
        verbose=not args.silent,
        thumb_res_lvl=args.thumb_res_lvl,
        yt_api_key=args.yt_api_key,
        disable_logs=args.disable_logs,
        log_duration=args.log_duration,
        reset=args.reset,
        not_cache=args.not_cache,
        window=args.window,
        min_duration=args.min_duration,
        keyword_limit=args.keyword_limit,
        keyword_filters=args.keyword_filters,
    )

    with analyser:
        fetched, read, analysed = False, False, False
        def fetch_data():
            nonlocal fetched
            if fetched:
                return
            if not analyser.is_cached:
                analyser.collect_data()
            analyser.enforce_integrity()
            fetched = True
        def read_data():
            fetch_data()
            nonlocal read
            if read:
                return
            analyser.read_data()
            analyser.refine_data()
            analyser.fetch_missing_messages()
            read = True
        def analyse_data():
            read_data()
            print()
            nonlocal analysed
            if analysed:
                return
            analyser.analyse_data()
            analysed = True
        
        # handle arguments
        if args.find_messages=="":
            read_data()
            for msg in analyser.messages:
                print(msg)
        elif args.find_messages:
            read_data()
            for msg in analyser.find_messages(
                args.find_messages, args.exact, not args.strict_case
            ):
                print(msg)

        if args.find_user_messages:
            read_data()
            messages=analyser.find_user_messages(id=args.user_id, username=args.username)
            if not messages:
                print("User not found")
            else:
                for msg in messages:
                    print(msg)

        if args.highlight_output:
            analyse_data()
            top_highlights = analyser.get_highlights(
                top=args.top,
                output_mode=args.highlight_output,
                include=args.include_context,
                exclude=args.exclude_context,
                intensity_filters=args.intensity_filters,
            )

        if args.open_in_chrome and top_highlights:
            analyse_data()
            for highlight in top_highlights:
                highlight.open_in_browser()

        if args.show_cached:
            print(analyser.cached_ids())

        if args.export:
            analyse_data()
            path = None if args.export == "default" else args.export
            analyser.export_data(args.export_folder_name, path, args.open_export_folder)

        if args.graph:
            analyse_data()
            analyser.show_graph()

        if args.wordcloud:
            analyse_data()
            analyser.generate_wordcloud(scale=args.wordcloud_scale).to_image().show()

        if not args.no_sound:
            os.system("echo ")  # windows notification sound


if __name__ == "__main__":
    main()



import streamanalyser as sa

# Find highlights of streams that are cached
if __name__ == "__main__":

    # Limited to merely 1000 messages for this example as it'd take way longer to fetch all messages.
    message_limit = 1000

    # Creating an empty analyser object instance to get the cached ids
    for id in sa.StreamAnalyser("").filehandler.get_cached_ids():
        with sa.StreamAnalyser(id, msglimit=message_limit) as analyser:
            analyser.analyse()
            analyser.get_highlights(
                top=len(analyser.highlights), output_mode="detailed"
            )
            print(f"Analysed {id} ({analyser.metadata['title']})")



from collections import Counter
from datetime import timedelta
import os
import random
import traceback
from shutil import copyfile
from typing import Tuple
from time import time
from colorama.ansi import Back, Style

from wordcloud import WordCloud

from .modules import (
    loggersetup,
    filehandler,
    datacollector,
    datarefiner,
    chatanalyser,
    structures,
    utils,
    cli,
)

DIR_PATH = os.path.dirname(os.path.realpath(__file__))
DEFAULT_CONTEXT_SOURCE_PATH = os.path.join(DIR_PATH, "data", "default_contexts.json")
DEFAULT_FONT_PATH = os.path.join(DIR_PATH, "fonts", "NotoSansCJKjp-Bold.ttf")
with open(rf"{DIR_PATH}/data/keyword_filters.txt", "r", encoding="utf-8") as f:
    DEFAULT_KEYWORD_FILTERS = [kw.strip("\n") for kw in f.readlines()]


class StreamAnalyser:
    """A class that analyses live streams.

    Args:
        sid (str): Video id of the stream.

        msglimit (int|None, optional): Message amount to fetch. Defaults to None.

        verbose (bool, optional): Print output to console. Defaults to False.

        thumb_res_lvl (int, optional): Resolution level of the thumbnail.
            Resolution levels are sorted by quality with 0 being lowest
            and 3 being highest. Defaults to 2.

        yt_api_key (str, optional): YouTube API key to provide full functionality.
            The class would still work 99% even if the api key is not provided, So
            it is not necessary, but still recommended for edge cases.
            Defaults to None.

        disable_logs (bool, optional): Disable logging.
            Log files' location can be found in `filehandler` module.
            Defaults to False.

        keep_logs (bool, optional): Do not delete logs. See `log_duration`
            for more information. Defaults to False.

        log_duration (int, optional): Log duration in days. Logs that are
            older than this value will be deleted if `keep_logs` option is
            False. Cache of the current week can only be deleted after 7 days.
            Defaults to 15.

        storage_path (str, optional): Folder to store files related to this
            module. Defaults to "C:\\Stream Analyser"

        reset (bool, optional): Clear existing cache when initializing the object.
            Defaults to False

        not_cache(bool, optional): Not cache stream data when using
            `analyse` function. Isn't recommended to use since fetching
            live chat messages takes quite a lot of time. The mere use-case
            is if the data will be used only for once. Even so, the cache
            can be deleted later on with the `clear_cache` function. Defaults
            to False.

        keep_cache (bool, optional): Do not delete any of the cached
            files, anytime (except the current session's cache if
            `not_cache` is True). Isn't recommended to use since lots
            of cached data might take up unnecessary space if not
            handled. If set to False, cached data will be deleted
            occasionally according to the chosen algorithm. See
            `cache_deletion_algorithm` for more info. Defaults to False.

        cache_deletion_algorithm (str, optional): Algorithm to delete
            cached files. Options are as follows:
                - lru (Least recently used): Deletes least recently
                    used cache. (recommended)
                - mru (Most recently used): Deletes most recently
                    used cache.
                - fifo (First in first out): Deletes oldest cache.
                - rr (Random replacement): Deletes random cache. (uhh...)
            Defaults to 'lru'.

        cache_limit (int, optional): Cache file amount to keep. Cached
            files will be deleted if file amount exceeds this value
            according to `cache_deletion_algorithm` option. Defaults to 50.

        min_duration (int): Minimum highlight duration (in seconds) to detect.
            Defaults to 5

        window (int, optional): Time interval to calculate moving averages.
            Defaults to 30.

        threshold_constant(int, optional): The value that divides average
            highlight duration. Defaults to 3.

        keyword_limit(int, optional): Keyword amount to retrieve. Defaults to 4.

        keyword_filters(list, optional): Keywords to filter. Defaults to [].

        intensity_levels (list[str], optional): See `init_intensity` function in
            `chat_analyser` module for information. Defaults to [].

        intensity_constants (list[str], optional): See `init_intensity` function in
            `chat_analyser` module for information. Defaults to [].

        intensity_colors (list[str], optional): See `init_intensity` function in
            `chat_analyser` module for information. Defaults to [].

        keep_analysis_data (bool, optional): Keeps analysis data such as message
            frequency, moving average and annotations until the object terminates.
            Set to False to conserve more memory during runtime if the data won't be
            needed. Defaults to True.

        default_context_path (str, optional): Default path to context json file.
            Defaults to DEFAULT_CONTEXT_SOURCE_PATH.

        stop_words_path (bool, optional): Default stop word file (.txt) path to exclude in
            keyphrase collocations. Defaults to None.
    """

    def __init__(
        self,
        sid,
        msglimit=None,
        verbose=False,
        thumb_res_lvl=2,
        yt_api_key=None,
        disable_logs=False,
        keep_logs=False,
        log_duration=15,
        storage_path="C:\\Stream Analyser",
        reset=False,
        not_cache=False,
        keep_cache=False,
        cache_deletion_algorithm="lru",
        cache_limit=50,
        min_duration=15,
        window=30,
        threshold_constant=3,
        keyword_limit=4,
        keyword_filters=[],
        intensity_levels=[],
        intensity_constants=[],
        intensity_colors=[],
        keep_analysis_data=True,
        default_context_path=DEFAULT_CONTEXT_SOURCE_PATH,
        stop_words_path = None
    ):

        self.sid = sid
        self.msglimit = msglimit
        self.verbose = verbose
        self.thumb_res_lvl = thumb_res_lvl
        self.yt_api_key = yt_api_key
        self.disable_logs = disable_logs
        self.not_cache = not_cache
        self.keep_cache = keep_cache
        self.cache_limit = cache_limit
        self.min_duration = min_duration
        self.window = window
        self.threshold_constant = threshold_constant
        self.keyword_limit = keyword_limit
        self.keyword_filters = keyword_filters + DEFAULT_KEYWORD_FILTERS
        self.intensity_levels = intensity_levels
        self.intensity_constants = intensity_constants
        self.intensity_colors = intensity_colors
        self.keep_analysis_data = keep_analysis_data
        self.default_context_path = default_context_path
        self.stop_words_path = stop_words_path

        self._raw_messages = {}
        self.messages = []
        self.authors = []
        self.highlights = []
        self.wordcloud = None
        self.fig = None
        self.metadata = {}
        self.context_source = structures.ContextSourceManager([])

        self.filehandler = filehandler.FileHandler(storage_path=storage_path)
        self.logger = loggersetup.create_logger(__file__, self.filehandler.log_path, sid=sid)
        self.collector = datacollector.DataCollector(sid, log_path=self.filehandler.log_path, msglimit=msglimit, verbose=verbose, yt_api_key=yt_api_key)
        self.refiner = datarefiner.DataRefiner(log_path=self.filehandler.log_path, verbose=verbose)
        self.canalyser = None  # It's recommended to empty this variable by hand to conserve memory after using the analysis data. See `keep_analysis_data` option for more.

        if disable_logs:
            self._disable_logs()

        if not keep_logs:
            self.filehandler._delete_old_files(self.filehandler.log_path, log_duration)
        self.logger.info(("=" * 100) + "=" * 15)
        self.logger.info(f"{'='*20} SESSION START {'='*80}")
        self.logger.info(("=" * 100) + "=" * 15)
        
        self.logger.info("Analyser initiated with following parameters:")
        self.logger.debug(f"{msglimit=}")
        self.logger.debug(f"{verbose=}")
        self.logger.debug(f"{thumb_res_lvl=}")
        self.logger.debug(f"{yt_api_key=}")
        self.logger.debug(f"{disable_logs=}")
        self.logger.debug(f"{keep_logs=}")
        self.logger.debug(f"{log_duration=}")
        self.logger.debug(f"{storage_path=}")
        self.logger.debug(f"{reset=}")
        self.logger.debug(f"{not_cache=}")
        self.logger.debug(f"{keep_cache=}")
        self.logger.debug(f"{cache_deletion_algorithm=}")
        self.logger.debug(f"{cache_limit=}")
        self.logger.debug(f"{min_duration=}")
        self.logger.debug(f"{window=}")
        self.logger.debug(f"{threshold_constant=}")
        self.logger.debug(f"{keyword_limit=}")
        self.logger.debug(f"{keyword_filters=}")
        self.logger.debug(f"{intensity_levels=}")
        self.logger.debug(f"{intensity_constants=}")
        self.logger.debug(f"{intensity_colors=}")
        self.logger.debug(f"{keep_analysis_data=}")
        self.logger.debug(f"{default_context_path=}")
        self.logger.debug(f"{stop_words_path=}")


        self.filehandler.create_cache_dir(self.sid)
        if reset:
            self.clear_cache(delete_root_folder=False)

        if not keep_cache:
            famount = self.filehandler.dir_amount(self.filehandler.cache_path)
            if cache_limit < 1:
                raise ValueError("Cache limit must be a natural number")
            if famount > cache_limit:
                self.logger.warning(
                    f"Cache limit has been exceeded by {famount-cache_limit}"
                )
            while famount > cache_limit:
                self.clear_cache(cache_deletion_algorithm)
                famount -= 1

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if exc_type is not None:
            traceback.print_exception(exc_type, exc_value, tb)
        if self.not_cache:
            self.clear_cache()

    def _disable_logs(self):
        self.logger.disabled = True
        self.filehandler.logger.disabled = True
        self.collector.logger.disabled = True
        self.refiner.logger.disabled = True

    def _cache_metadata(self, metadata):
        self.filehandler.cache_metadata(metadata)

    def _cache_messages(self, raw_messages):
        self.filehandler.cache_messages(raw_messages)

    def clear_cache(self, cache_deletion_algorithm=None, delete_root_folder=True):
        self.filehandler.clear_cache(cache_deletion_algorithm, delete_root_folder)

    def collect_data(self):
        """Collects and caches stream data:
        - messages
        - metadata (title, channel, duration etc.)
        """
        # collect data
        metadata = self.collector.collect_metadata()
        raw_messages = self.collector.fetch_raw_messages()

        # cache data
        self._cache_metadata(metadata)
        self._cache_messages(raw_messages)

    def read_data(self):
        """Reads cached data"""
        if self.verbose:
            print("Reading messages...", end="\r")

        self._raw_messages = self.filehandler.read_messages()
        self.update_metadata(self.filehandler.read_metadata())

        if "is-complete" in self.metadata.keys():
            if not self.metadata["is-complete"]:
                self.update_metadata({"is-complete": self.collector.iscomplete})
        else:
            self.update_metadata({"is-complete": self.collector.iscomplete})

        if self.verbose:
            print("Reading messages... done")

    def refine_data(self):
        """Refines read data"""
        self.messages = self.refiner.refine_raw_messages(
            self._raw_messages, self.msglimit
        )
        # we don't need raw messages anymore
        # empty them so they don't take up space
        self._raw_messages = None
        self.authors = self.refiner.get_authors()

    def analyse_data(self):
        """Analyses refined data and detects highligths"""
        self.canalyser = chatanalyser.ChatAnalyser(
            log_path=self.filehandler.log_path,
            refined_messages=self.messages,
            default_context_path=self.default_context_path,
            stream_id=self.sid,
            verbose=self.verbose,
            keyword_filters=self.keyword_filters,
            keyword_limit=self.keyword_limit,
            min_duration=self.min_duration,
            threshold_constant=self.threshold_constant,
            window=self.window,
            stop_words_path=self.stop_words_path
        )
        if self.disable_logs:
            self.canalyser.logger.disabled = True
        self.canalyser.source.paths.extend(self.context_source.paths)

        self.canalyser.analyse(
            levels=self.intensity_levels,
            constants=self.intensity_constants,
            colors=self.intensity_colors,
            autofix_context_collision=True
        )
        self.highlights = self.canalyser.highlights

        if not self.keep_analysis_data:
            self.canalyser = None

    def analyse(self):
        if not self.is_cached:
            self.collect_data()
        self.read_data()
        self.refine_data()
        self.enforce_integrity()
        self.fetch_missing_messages()
        self.analyse_data()

    def _check_integrity(self, autofix=False) -> Tuple[list, list]:
        return self.filehandler.check_integrity(autofix=autofix)

    @property
    def is_cached(self):
        return self.filehandler.is_cached()

    def fetch_missing_messages(self):
        """Checks and fetches missing messages if there's any.
        Should be used when the message limit is increased or
        set to None. It also caches, reads and refines the
        messages by itself as well.

        For instance if 1000 messages had been fetched before,
        and the user is requesting 1200 messages now, the function
        will only fetch the last 200 messages instead of starting
        all over again.
        """

        self.logger.info("Checking missing messages")

        if "is-complete" not in self.metadata.keys():
            self.logger.debug(
                "Could not fetch missing messages since messages are not collected yet"
            )
            return

        if self.metadata["is-complete"]:
            self.logger.debug("Messages are already complete")
            return
        if self.verbose:
            print("Checking missing messages...", end="\r")

        raw_messages = self.filehandler.read_messages()
        last_time = raw_messages[-1]["time_in_seconds"]
        current_amount = len(raw_messages)

        if not self.metadata["is-complete"] and not self.msglimit:
            target_amount = None
        else:
            target_amount = self.msglimit - current_amount
            if target_amount <= 0:
                self.logger.debug("No missing messages detected")
                if self.verbose:
                    print("Checking missing messages... done")
                return

        if self.verbose:
            print("Checking missing messages... done")

        missing_messages = self.collector.fetch_missing_messages(
            start_time=last_time,
            current_amount=current_amount,
            target_amount=target_amount,
        )
        self.filehandler.cache_messages(raw_messages + missing_messages)
        self.messages = self.messages + self.refiner.refine_raw_messages(
            missing_messages
        )
        self.authors = self.authors + self.refiner.get_authors()
        self.update_metadata({"is-complete": self.collector.iscomplete})

    def update_metadata(self, new_dict):
        """Updates both metadata file and variable"""
        self.metadata = {**self.metadata, **new_dict}
        self.filehandler.cache_metadata(self.metadata)
        self.logger.info("Updated metadata")

    def enforce_integrity(self):
        """Enforces file integrity by recollecting missing
        data and deleting unnecessary cache files"""

        missing_files, _ = self._check_integrity(autofix=True)
        for missing_file in missing_files:
            if missing_file == self.filehandler.message_fname + ".gz":
                self.logger.warning("Message file is missing")
                self.filehandler.cache_messages(self.collector.fetch_raw_messages())
            elif missing_file == self.filehandler.metadata_fname:
                self.logger.warning("Metadata file is missing")
                self.filehandler.cache_metadata(self.collector.collect_metadata())

        try:
            # TODO fix logic
            #self.fetch_missing_messages()
            pass
        except KeyError:
            pass
        except Exception as e:
            self.logger.error(e.__class__.__name__ + e)

    def generate_wordcloud(self, font_path=None, scale=3, background="aliceblue"):
        """Returns a basic word cloud

        Args:
            font_path (str, optional): Custom font path. Defaults to None.
            scale (int, optional): Scale of the resulting wordcloud.
                Might want to decrease it for more performance. Defaults to 3.
            background (str, optional): Background color. Defaults to "aliceblue".
        """

        self.logger.info("Generating word cloud")
        self.logger.debug(f"{font_path=}")
        self.logger.debug(f"{scale=}")
        self.logger.debug(f"{background=}")

        if self.verbose:
            print("Generating word cloud...", end="\r")

        if not font_path:
            font_path = DEFAULT_FONT_PATH

        # get all words from the chat
        wordlist = [msg.text.replace("_", "") for msg in self.messages]

        # shuffle word list to minimize the issue where repeating
        # consecutive messages merge together in the word cloud
        random.shuffle(wordlist)

        try:
            word_cloud = WordCloud(
                font_path=font_path,
                scale=scale,
                background_color=background,
            ).generate(" ".join(wordlist))
        except Exception as e:
            self.logger.error(f"Could not generate wordcloud: {e}")
            if self.verbose:
                print("Generating word cloud... error")
            raise e

        if self.verbose:
            print("Generating word cloud... done")
        self.wordcloud = word_cloud
        return self.wordcloud

    def find_messages(
        self, search_phrase, exact=False, ignore_case=True
    ) -> list[structures.Message, structures.Superchat, structures.Membership]:
        """Finds messages containing a specific phrase or exactly the same phrase.

        Args:
            search_phrase (str): The phrase to search.
            exact (bool, optional): If the phrase has to be exactly the same or not. Defaults to False.
            ignore_case (bool, optional): Ignore letter cases. Defaults to True.

        Returns:
            list[Message,Superchat,Membership]: List of messages that contains the given phrase.
        """

        messages_to_return = []
        self.logger.info("Finding messages")
        self.logger.debug(f"{search_phrase=}")
        self.logger.debug(f"{exact=}")
        self.logger.debug(f"{ignore_case=}")

        if ignore_case:
            search_phrase = search_phrase.lower()

        tmp_message = None
        for original_message in self.messages:
            tmp_message = original_message.text
            if ignore_case:
                tmp_message = tmp_message.lower()
            if (exact and search_phrase == tmp_message) or (
                not exact and search_phrase in tmp_message
            ):
                messages_to_return.append(original_message)
        return messages_to_return

    def find_user_messages(self, username=None, id=None) -> list[structures.Message, structures.Superchat, structures.Membership]:
        """Finds messages by either username or user id.

        Args:
            username (str, optional): Target username. Defaults to None.
                Note that it's more reliable to use id since there might
                be same users with the same name or username changes.

            id (str, optional): Target user id. Defaults to None.

        Raises:
            ValueError: Should provide either username or id.

        Returns:
            list[Message,Membership,Superchat]: Messages the user typed.
        """

        self.logger.info("Finding user messages")
        self.logger.debug(f"{id=}")
        self.logger.debug(f"{username=}")

        if not username and not id:
            self.logger.error("Should provide either username or id")
            raise ValueError("Should provide either username or id.")
        if username and id:
            self.logger.warning("Should only provide one argument. Moving on with id.")
            username = None

        messages_to_return = []
        for message in self.messages:
            if (id and message.author.id == id) or (
                username and message.author.name == username
            ):
                messages_to_return.append(message)

        return messages_to_return

    def most_used_phrase(self, exclude=[], normalize=True) -> Tuple[str, int]:
        """Returns most frequently used phrase

        Args:
            exclude (list, optional): List of words to exclude from the search.
                Defaults to [].

            normalize (bool, optional): Option for the word be normalized
                to cover more instances. Defaults to True.

        Returns:
            Tuple[str, int]: Most used word and it's frequency
        """

        # return "草"    # would probably work lol

        self.logger.info("Finding most used word")
        self.logger.debug(f"{exclude=}")
        self.logger.debug(f"{normalize=}")

        if isinstance(exclude, str):
            exclude = list(exclude)

        words = []
        for message in self.messages:
            words.extend(message.text.split(" "))

        if normalize:
            words = [utils.normalize(word) for word in words]

        idx = 0
        while exclude and Counter(words).most_common(1 + idx)[idx][0] in exclude:
            idx += 1
            if idx == len(words) - 1:
                return Counter(words).most_common(1 + idx)[idx]

        self.logger.debug(f"Most used phrase: {Counter(words).most_common(1+idx)[idx]}")
        return Counter(words).most_common(1 + idx)[idx]

    @property
    def total_message_amount(self):
        return len(self.messages)

    def export_data(self, folder_name=None, path=None, open_folder=False):
        """Exports the analysed data to the path.
        Exported data are:
        - messages
        - highlights
        - metadata
        - thumbnail
        - graph
        - word cloud

        Args:
            folder_name (str|None, optional): File name to export the results. Defaults to None,
                which exports data under the file named current UNIX timestamp.
            path (str|None, optional): Path to export the results. Defaults to None,
                which exports data to the default path.
            open_folder (bool, optional): Open the export folder in file explorer
                after exporting. Defaults to false

        """

        self.logger.info("Exporting data")
        self.logger.debug(f"{folder_name=}")
        self.logger.debug(f"{path=}")

        if self.verbose:
            print("Exporting data...", end="\r")

        if not path:
            path = self.filehandler.export_path

        try:
            self.filehandler.create_dir_if_not_exists(path)
        except PermissionError as pe:
            self.logger.error(pe)
            return

        if not folder_name:
            folder_name = str(int(time()))
        else:
            # check if there's already a file with the same name
            # if so, add UNIX timestamp to make it unique
            for name in os.listdir(path):
                if os.path.isdir(os.path.join(path, name)) and name == folder_name:
                    folder_name = folder_name + "_" + str(int(time()))
                    warn_msg = f"{name} already exists, renaming to {folder_name}"
                    self.logger.warning(warn_msg)
                    break

        target_path = path + "\\" + folder_name
        self.filehandler.create_dir_if_not_exists(target_path)

        # export messages
        self.filehandler._decompress_file(
            os.path.join(self.filehandler.sid_path, self.filehandler.message_fname)
        )
        copyfile(
            src=os.path.join(self.filehandler.sid_path, self.filehandler.message_fname),
            dst=os.path.join(target_path, self.filehandler.message_fname),
        )
        self.filehandler._compress_file(
            os.path.join(self.filehandler.sid_path, self.filehandler.message_fname)
        )
        self.logger.info("Exported messages")

        # export metadata
        copyfile(
            src=os.path.join(
                self.filehandler.sid_path, self.filehandler.metadata_fname
            ),
            dst=os.path.join(target_path, self.filehandler.metadata_fname),
        )
        self.logger.info("Exported metadata")

        # export thumbnail
        self.filehandler.download_thumbnail(
            self.collector.get_thumbnail_url(self.thumb_res_lvl),
            os.path.join(target_path, self.filehandler.thumbnail_fname),
        )
        self.logger.info("Exported thumbnail")

        # export word cloud
        if self.messages:
            self.generate_wordcloud().to_file(
                os.path.join(target_path, "wordcloud.jpg")
            )
            self.logger.info("Exported wordcloud")

        # export highlights
        if self.highlights:
            hl_path = os.path.join(target_path, "highlights.txt")
            with open(hl_path, "w", encoding="utf-8") as file:
                file.writelines([hl.colorless_str + "\n" for hl in self.highlights])
            self.logger.info("Exported highlights")

        if self.fig:
            self.fig.savefig(os.path.join(target_path, "graph.png"))
            self.logger.info("Exported graph")

        if open_folder:
            try:
                os.startfile(target_path)
                self.logger.info(f"Opened {target_path} in file explorer")
            except FileNotFoundError:
                self.logger.error(f"Couldn't find {target_path}")

        if self.verbose:
            print("Exporting data... done")

    def get_highlights(
        self, top=None, output_mode=None, include=[], exclude=[], intensity_filters=[]
    ) -> list[structures.Highlight]:
        """A method to return filtered highlights.

        Args:
            top (int, optional): Top n highlights to fetch, sorted by fdelta attribute.
                Defaults to None, which returns all.

            output_mode (str, optional): Mode to print output of the highlights on
                the console. Options are:
                    - 'detailed': Show details of the highlights.
                    - 'summary': Show only time and intensity of the highlights
                    - 'url': Show only time and url of the highlights
                    - None: No output.
                Defaults to None.

            include (list[str]|str, optional): List of reactions to see. Defaults to [].
                Reaction names can be found in `data\context.json`.

            exclude (list[str]|str, optional): List of reactions to not see.
                Overrides include. Defaults to [].
                Reaction names can be found in `data\context.json`.

            intensity_filters (list[str]|str, optional): Intensity levels to filter out.
                Defaults to [].

        Returns:
            list[Highlight]: List of highlights
        """

        self.logger.info("Printing highlights")
        self.logger.debug(f"{top=}")
        self.logger.debug(f"{output_mode=}")
        self.logger.debug(f"{include=}")
        self.logger.debug(f"{exclude=}")
        self.logger.debug(f"{intensity_filters=}")

        if isinstance(include, str):
            include = list(include)
        if isinstance(exclude, str):
            exclude = list(exclude)
        if isinstance(intensity_filters, str):
            intensity_filters = list(intensity_filters)

        if top and top < 0:
            self.logger.error("Top value cannot be negative")
            raise ValueError("Top value cannot be negative.")

        count = 0
        highlights_to_return = []
        if top:
            highlights = sorted(self.highlights, key=lambda x: x.fdelta, reverse=True)
        else:
            highlights = self.highlights

        if output_mode:
            if output_mode == "detailed":
                title = "Highlights:"
            elif output_mode == "url":
                title = "Links:"
            elif output_mode == "summary":
                title = "Summary:"
            else:
                self.logger.error("Invalid output_mode")
                raise ValueError("Invalid output_mode")
            print("\n" + Back.RED + title + Style.RESET_ALL)
        for highlight in highlights:
            skip = False
            for context in highlight.contexts:
                if exclude and context in exclude:
                    skip = True
                    break
            if skip:
                continue
            for context in highlight.contexts:
                if (include and context in include) or not include:
                    if not highlight.intensity.level in intensity_filters:
                        if output_mode == "detailed":
                            print(highlight)
                        elif output_mode == "summary":
                            print(
                                "{}: {}".format(
                                    timedelta(seconds=int(highlight.time)),
                                    highlight.intensity.colored_level,
                                )
                            )
                        elif output_mode == "url":
                            print(
                                highlight.intensity.color
                                + str(timedelta(seconds=int(highlight.time)))
                                + Style.RESET_ALL,
                                "->",
                                highlight.url,
                            )
                        elif output_mode is None:
                            pass
                        else:
                            self.logger.error("Invalid output mode")
                            raise ValueError(f'Invalid output mode: "{output_mode}"')
                        highlights_to_return.append(highlight)
                        self.logger.debug(highlight.colorless_str)
                        count += 1
                        if top and count == top:
                            return highlights_to_return
                        break

        return highlights_to_return

    def add_context(self, reaction_to, phrases):
        self.filehandler.add_context(reaction_to, phrases)

    def remove_context(self, reaction_to):
        self.filehandler.remove_context(reaction_to)

    def cached_ids(self) -> list[str]:
        return self.filehandler.get_cached_ids()

    def open_cache_folder(self):
        self.filehandler.open_cache_folder(self.sid)

    def show_graph(self):
        self.canalyser.draw_graph(self.metadata["title"])
        self.fig = self.canalyser.fig
        self.fig.show()


if __name__ == "__main__":
    cli.main()



import unittest
import warnings

from modules.datacollector import DataCollector


# TODO Fix:
# On: Chatdownloader call with limit
# Warning: unclosed <ssl.SSLSocket> on break


class TestDataCollector(unittest.TestCase):
    def test_collect_metadata(self):
        metadata = self.collector.collect_metadata()
        self.assertEqual(
            metadata["title"], "4 A.M Study Session 📚 - [lofi hip hop/chill beats]"
        )

    def test_fetch_raw_messages(self):
        raw_messages = self.collector.fetch_raw_messages()
        self.assertEqual(
            raw_messages[0]["message_id"],
            "CkUKGkNOU255NHJNblBBQ0ZSaXN3UW9kQkdjQWVBEidDT1My"
            + "dV92TG5QQUNGY3RsbXdvZDVaZ01RUTE2MTk0NjM2MTQzNTE%3D",
        )

        # messages fetched with a limit
        # should be considered incomplete
        self.assertFalse(self.collector.iscomplete)

    def test_fetch_missing_messages(self):
        raw_messages = self.collector.fetch_missing_messages(20, 10, 2)

        self.assertEqual(raw_messages[0]["author"]["id"], "UCjbWfSrdmmtdRSRYcoqjc-Q")

        # messages fetched with a limit
        # should be considered incomplete
        self.assertFalse(self.collector.iscomplete)

    def setUp(self):
        warnings.simplefilter("ignore", category=ResourceWarning)
        self.example_url = "TURbeWK2wwg"
        self.collector = DataCollector(self.example_url, msglimit=1, log_path=None)
        self.collector.logger.disabled = True

    def tearDown(self):
        del self.collector


if __name__ == "__main__":
    unittest.main()



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_desc = fh.read()

with open("requirements.txt") as f:
    requirements = [rq.replace("==", ">=") for rq in f.read().splitlines()]

setup(
    name="stream-analyser",
    version="v0.3.2",
    author="emso-c",
    author_email="emsoc192@gmail.com",
    description=("A tool that analyses live streams"),
    keywords=["youtube", "live", "stream", "chat", "highlight", "analysis"],
    url="https://github.com/emso-c/stream-analyser",
    long_description=long_desc,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: Microsoft :: Windows :: Windows 10",
        "Environment :: Console",
        "Natural Language :: English",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Information Analysis",
    ],
    entry_points={'console_scripts':[
        'streamanalyser = streamanalyser.modules.cli:main'
    ]},
    packages=find_packages(),
    package_dir={"": "."},
    python_requires=">=3.7",
    include_package_data=True,
    package_data={'streamanalyser': [
        'data/default_contexts.json',
        'data/keyword_filters.txt',
        'data/monogram_stop_punctuations.txt',
        'data/monogram_stop_words.txt',
        'fonts/NotoSansCJKjp-Bold.ttf',
    ]},
    install_requires=requirements,
)



import logging
import os
import unittest

from modules.loggersetup import create_logger



class TestLoggerSetup(unittest.TestCase):
    def test_createlogger(self):
        logger = create_logger(
            __file__,
            folder_path="C:\\Stream Analyser\\Logs",
            file_name="_test.log",
            sid=None,
            format="%(module)s:%(levelname)s:%(message)s",
            mode="w",
        )
        self.assertTrue(os.path.exists(self.test_log_path))

        logger.info("test")
        logger.debug("test")
        logger.warning("test")
        # logger.error('test')
        # logger.critical('test')
        lvls = ["INFO", "DEBUG", "WARNING"]  #'ERROR','CRITICAL']
        with open(self.test_log_path, "r") as f:
            for i, line in enumerate(f.readlines()):
                self.assertEqual(line, f"test_loggersetup:{lvls[i]}:test\n")
    def setUp(self):
        self.test_log_path = "C:\\Stream Analyser\\Logs\\_test.log"
    def tearDown(self):
        logging.shutdown()
        os.remove(self.test_log_path)


if __name__ == "__main__":
    unittest.main()



import streamanalyser as sa

# Find instances of a specified message that has been written by the same user.
if __name__ == "__main__":
    analyser = sa.StreamAnalyser(
        "pVRvx4FBEwU", msglimit=1000, verbose=True, disable_logs=True, not_cache=True
    )

    with analyser:
        analyser.collect_data()
        analyser.read_data()
        analyser.refine_data()

        message_to_search = ":_koroneIiyubi::_koroneIiyubi::_koroneIiyubi:"
        user_name = "Rarely Spotted"

        def solution1():
            found_messages = analyser.find_messages(
                message_to_search,
                exact=True,
                ignore_case=False,
            )
            for msg in found_messages:
                if msg.author.name == user_name:
                    yield msg

        def solution2():
            for msg in analyser.find_user_messages(username=user_name):
                if msg.text == message_to_search:
                    yield msg

        print("Solution 1")
        [print(msg) for msg in solution1()]
        print("\nSolution 2")
        [print(msg) for msg in solution2()]



import nltk
import re
import string
import os

PHRASE_END_DELIMITER = "%phrase_end%" # to not mix continous chat messages

NLTK_DATA_PATH = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "..", "data", "nltk_data"
)
MONOGRAM_STOP_PUNCTUATIONS_PATH = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "..", "data", "monogram_stop_punctuations.txt"
)
MONOGRAM_STOP_WORDS_PATH = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "..", "data", "monogram_stop_words.txt"
)


class KeyphraseFinder:
    """A class that finds keyphrases from a live chat using natural language processing
    """
    # TODO more docs

    def __init__(
            self,
            chat,
            monogram_stop_words_path = MONOGRAM_STOP_WORDS_PATH,
            monogram_stop_punctuations_path = MONOGRAM_STOP_PUNCTUATIONS_PATH,
            stop_words_path = None,  # your custom stop words to exclude in collocations
            fix_phrases = [],
            punctuation_list=list(string.punctuation),
        ):

        nltk.download('punkt', download_dir=NLTK_DATA_PATH, quiet=True) 
        nltk.data.path = [NLTK_DATA_PATH]
        
        self.chat = chat
        self.fix_phrases = fix_phrases
        self.punctuation_list = punctuation_list
        self.stop_words_path = stop_words_path

        self.stop_words = None

        # TODO logger

        try:
            # always left an empty whitespace at the end of the txt file
            # get text data line by line
            with open(monogram_stop_words_path, 'r', encoding="utf-8") as file:
                self.monogram_stop_punctuations = [r[:-1] for r in file.readlines()]
            with open(monogram_stop_punctuations_path, 'r', encoding="utf-8") as file:
                self.monogram_stop_words = [r[:-1] for r in file.readlines()]
            if stop_words_path:
                with open(stop_words_path, 'r', encoding="utf-8") as file:
                    self.stop_words = [r[:-1] for r in file.readlines()]
        except FileNotFoundError as e:
            # log
            raise e

        self.monogram_stop_words_all = self.monogram_stop_words + self.monogram_stop_punctuations

    @staticmethod
    def _regex_partition(string, regex):
        if not re.search(regex, string):
            string = string.lower()

        first = re.split(regex, string, 1)[0]
        try:
            last = re.split(regex, string, 1)[1]
        except IndexError:
            last = ''

        regp = re.compile(regex)
        result = regp.search(string)

        try:
            match = result.group()
        except AttributeError:
            match = ''

        return first, match, last

    @staticmethod
    def _tokenize_yt_chat_message(string):
        first, sub_string, last = KeyphraseFinder._regex_partition(string, ":.+:")

        if sub_string:
            tokens = KeyphraseFinder._tokenize_yt_chat_message(last)
            return nltk.word_tokenize(first) + [sub_string] + tokens
        else:
            return nltk.word_tokenize(first)

    def _fix_token(self, token):
        for fix_phrase in self.fix_phrases:
            token = token.replace(fix_phrase[0], fix_phrase[1])
        return token

    def _merge_punctuations(self):
        offset = -1
        for i, token in enumerate(self.tokens):
            if i == len(self.tokens)-1 and offset != -1:
                self.tokens[offset:i+1] = [''.join(self.tokens[offset:i+1])]
                break
            if offset != -1 and token not in self.punctuation_list:
                self.tokens[offset:i] = [''.join(self.tokens[offset:i])]
                offset = -1
            if offset == -1 and token in self.punctuation_list:
                offset = i

    def _adjust_tokens(self):
        if len(self.tokens) == 1 and all([1 if ch in self.punctuation_list else 0 for ch in self.tokens[0]]):
            return
        for i, token in enumerate(self.tokens):
            if len(token) == 1 and all([1 if ch in self.punctuation_list else 0 for ch in token]):
                del self.tokens[i]

    def _tokenize_yt_chat(self):
        self.tokens=[]
        for yt_tokens in [KeyphraseFinder._tokenize_yt_chat_message(msg.text) for msg in self.chat]:
            self._merge_punctuations()
            self._adjust_tokens()
            self.tokens.extend(
                [token for token in list(yt_tokens)]+
                [PHRASE_END_DELIMITER]
            )
        return self.tokens

    @staticmethod
    def _configure_keyphrase_amount(max_keyphrase_amount, min_keyphrase_amount, chat_len, stoppers):
        stoppers = sorted(stoppers, reverse=True)
        for stopper in stoppers:
            if  chat_len < stopper:
                max_keyphrase_amount -= 1
        if max_keyphrase_amount<min_keyphrase_amount:
            max_keyphrase_amount = min_keyphrase_amount
        return max_keyphrase_amount, min_keyphrase_amount

    def ngram_keyphrase_analysis(self, max_ngram_size=7, max_keyphrase_amount=5, min_keyphrase_amount=2, replace_by_weight_score=False, keyphrase_per_ngram=20, min_ngram_size=1, stoppers=[100, 40, 10]) -> list:
        """Returns the most frequently used keyphrases using ngram frequency

            Searches through all ngrams from `max_ngram_size` to `min_ngram_size` to find
            collocations.

            Keyphrases found in ngrams with smaller sizes are overridden by their
            larger ngram counterparts since it's more likely that they contain less
            context.
                Example:
                    Assuming "happy new year" (ngram size is 3) keyphrase is already
                    found, keyphrases such as "new", "happy", "happy new" or "new year"
                    will be excluded from the results since a super-collocation that includes
                    all of them already exists.

            Result frequencies are sorted by their weighted score which is calculated
            with `frequency * ngram_size` formula as of now.

            Returned amount of keyphrases can differ between min and max sizes depending
            on the message amount (lower message amount means less keyphrases).
    
        Args:
            max_ngram_size (int, optional): Max ngram size to search for collocations.
                Lower this value to improve performance. Defaults to 7.
            max_keyphrase_amount (int, optional): Max keyphrase amount to return.
                Defaults to 5.
            min_keyphrase_amount (int, optional): Min keyphrase amount to return. Defaults to 2.
            replace_by_weight_score (bool, optional): [Experimental feature].
                If set to true, ngrams with smaller sizes will replace with their
                super-collocation if they outscore their super-collocation. Defaults to False.
                Example:
                    Considering following keyphrases:
                        "Hello there" (fre: 3, wscore: 6)
                        "Hello" (fre: 21, wscore: 21)
                    Since 21 > 6, "Hello there" will be replaced with "Hello", ignoring
                    "super-ngram contains more context" philosophy.
            keyphrase_per_ngram (int, optional): Top n keyphrases to return per ngram. Defaults to 20.
            min_ngram_size (int, optional): Minimum ngram size to search for collocations.
                Recommened value is 1. Defaults to 1.
            stoppers (int, optional): Fixed message amount points to lower returned keyphrase amount.
                Defaults to [100, 40, 10].

        Returns:
            list: List of most frequent keyphrases
        """

        # TODO parameter validation
        self._tokenize_yt_chat()

        max_keyphrase_amount, min_keyphrase_amount = KeyphraseFinder._configure_keyphrase_amount(
            max_keyphrase_amount, min_keyphrase_amount, len(self.chat), stoppers
        )

        seen_tuples = []
        for ngram_size in range(max_ngram_size, min_ngram_size-1, -1):
            fre_dist_list = nltk.FreqDist(nltk.ngrams(self.tokens, ngram_size)).most_common(keyphrase_per_ngram)
            for fre_tup in fre_dist_list:
                message = self._fix_token(' '.join(fre_tup[0]))
                frequency = fre_tup[1]
                weighted_score = frequency*ngram_size
                if not message:
                    continue
                if frequency <= 1:
                    continue
                if PHRASE_END_DELIMITER in message:
                    continue
                if ngram_size == 1 and message in self.monogram_stop_words:
                    continue
                if ngram_size > 1 and self.stop_words:
                    if any(1 if stop_word in message else 0 for stop_word in self.stop_words):
                        continue
                # punctuation tokens should be considered seperately
                if all(1 if ch in self.punctuation_list else 0 for ch in message):
                    if any([1 for tup in seen_tuples if message == tup[0]]): # already seen
                        continue
                else:
                    skip = False
                    for tup in seen_tuples: 
                        if message in tup[0]: # already seen in higher ngrams
                            if not replace_by_weight_score:
                                skip = True
                                break
                            if weighted_score > tup[3]:
                                seen_tuples.remove(tup)
                            else:
                                skip = True
                            break
                    if skip:
                        continue
                seen_tuples.append((message, frequency, ngram_size, weighted_score)) # TODO keyphrase object

                if len(seen_tuples) == max_keyphrase_amount or ngram_size == min_ngram_size:
                    return sorted(seen_tuples, key=lambda x: x[3], reverse=True)
        return sorted(seen_tuples, key=lambda x: x[3], reverse=True)



import streamanalyser as sa


# Get n most frequently used words with a recursive approach
if __name__ == "__main__":
    analyser = sa.StreamAnalyser(
        "pVRvx4FBEwU", msglimit=1000, verbose=True, disable_logs=True, not_cache=True
    )

    with analyser:
        analyser.collect_data()
        analyser.read_data()
        analyser.refine_data()

        n = 6
        mfup = []
        while n:
            mfup.append(analyser.most_used_phrase(exclude=mfup)[0])
            n -= 1

        for item in mfup:
            print(item)



import streamanalyser as sa

# Set a custom storage path
if __name__ == "__main__":
    analyser_obj = sa.StreamAnalyser(
        "pVRvx4FBEwU",
        verbose=True,
        disable_logs=True,
        not_cache=True,
        msglimit=100,
        storage_path="C:\\Stream_Analyser_Example_Custom_Storage"
    )
    with analyser_obj as analyser:
        analyser.analyse()
        analyser.get_highlights(output_mode="detailed")
        analyser.open_cache_folder()